import pyfnntw;
import numpy as np;
from time import time

ND = 10**7
RUNS = 4
TRIALS = 10
WARMUP = 1

overall_build_time = 0
print("Warming up")
for trial in range(1-WARMUP,TRIALS+1):
    trial_build_time = 0
    for _ in range(RUNS):
        
        # Define data, query
        data = np.random.uniform(size=(ND, 3))

        # Build tree
        start = time()
        tree = pyfnntw.Tree(data, 32, 2)
        build_time = (time() - start) * 1000
        trial_build_time += build_time

        print(f"{build_time=:.2f} ms")

    if trial > 0:
        overall_build_time += trial_build_time
        trial_avg_build = trial_build_time / RUNS
        print(f"Trial {trial} results: {trial_avg_build=:.2f} ms")

overall_avg_build = overall_build_time / RUNS / TRIALS
print(f"Overall Results: {overall_avg_build=:.2f} ms")



from scipy.spatial import cKDTree as Tree
import matplotlib.pyplot as plt
import numpy as np

# Load Data
data = np.load("data.npy")
query = np.load("query.npy")
results = np.load("results.npy")
indices = np.load("indices.npy")

# Check shapes are what we expect
DIMS = 3
QUERY = 10**6
DATA = 10**5
assert data.shape == (DATA, DIMS), "incorrect data shape"
assert query.shape == (QUERY, DIMS), "incorrect query shape"
assert results.shape == (QUERY, ), "incorrect results shape"

# Construct and query tree
tree = Tree(data, leafsize=32, boxsize=1.0)
(expected, idxs) = tree.query(query, workers=-1)

# Check results
# TOL = 1e-6
correct1 = np.isclose(expected**2, results)
correct = (indices == idxs)
if np.all(correct) and np.all(correct1):
    print("Success")
else:
    num = np.invert(correct).sum()
    num1 = np.invert(correct1).sum()
    print(f"Failure: {num:,}/{num:,} of {len(query):,}")

# incorrect = []
# for i in range(len(query)):

#     if idxs[i] != indices[i]:
#         print(i, idxs[i], expected[i]**2, "!=", indices[i], results[i])
#         incorrect.append((i, idxs[i], indices[i]))
#         assert results[i] > expected[i]**2

#     # else:
#         # print(idxs[i], expected[i]**2, "==", indices[i], results[i])


# min = np.zeros(data.shape[1])
# max = np.ones(data.shape[1])

# # print(data)
# # print(query)

# split0 = 0.3289118679348304
# split1 = 0.49667444751900025


# plt.figure(figsize=(12,8))
# plt.plot(data[:,0], data[:,1], 'o', label="data")
# plt.plot(query[:,0], query[:,1], 'x', label="query")
# plt.plot([split0]*2, [0,1], 'k')
# plt.plot([split0, 1], [split1]*2, 'k')
# plt.legend()
# for (query_idx, right, wrong) in incorrect:
#     q = query[query_idx]
#     r = data[right]
#     w = data[wrong]
#     plt.plot([q[0], r[0]], [q[1], r[1]], "g")
#     plt.plot([q[0], w[0]], [q[1], w[1]], "r")
# plt.xlim(0,1)
# plt.ylim(0,1)
# plt.savefig("kdtree.png")


import numpy as np
from time import perf_counter as time
import pyfnntw
import os
os.environ.setdefault("TCMALLOC_LARGE_ALLOC_REPORT_THRESHOLD", f"{100*2**30}")


NRAND = 10**5
NQUERY = 10**7
RUNS = 10
WARMUP = 10
LS = 32
KK = [1, 32]

# Generate data 
DATA = np.random.uniform(size=(NRAND, 3))
QUERIES = np.random.uniform(size=(NQUERY, 3))

for lib in ["fnntw"]:
    print()
    print(f"{lib} results (d={NRAND:,}, q={NQUERY:,}, ls={LS}, runs={RUNS})")

    build_time = 0
    kdtree = None
    # Generate queries
    for run in range(-WARMUP, RUNS):
        # Build tree
        start = time()
        kdtree = pyfnntw.Treef64(DATA, leafsize=LS, par_split_level=3)
        bt = (time() - start)*1000
        if run > 0:
            build_time += bt
    avg_build = build_time / RUNS
    print(f"{lib}: {avg_build=:.3f} ms")

    for K in KK:
        query_time = 0
        qt = 0
        for run in range(-WARMUP, RUNS):

            # Query tree
            if lib == "scipy":
                start = time()
                dist, idx = kdtree.query(QUERIES, k=K, workers=-1)        
                qt = (time() - start)*1000
            elif lib == "pykdtree":
                start = time()
                dist, idx = kdtree.query(QUERIES, k=K, sqr_dists=False)
                qt = (time() - start)*1000
            elif lib == "fnntw":
                start = time()
                dist, idx = kdtree.query(QUERIES, K)
                qt = (time() - start)*1000
            else:
                assert False
            if run > 0:
                query_time += qt
            
            # Print run results
            # print(f"Run {run}: {bt=:.3f} ms; {qt=:.3f} ms")

        avg_query = query_time / RUNS
        print(f"{lib} k={K}: {avg_query=:.3f} ms")
    print()

print("periodic results")
for lib in ["fnntw"]:
    print()
    print(f"{lib} results (d={NRAND:,}, q={NQUERY:,}, ls={LS}, runs={RUNS})")
    for K in KK:
        query_time = 0
        build_time = 0
        kdtree = None
        # Generate queries
        QUERIES = np.random.uniform(size=(NQUERY, 3))
        kdtree = pyfnntw.Treef64(DATA, boxsize = np.array([1.0]*3), leafsize=LS, par_split_level=3)

        for run in range(RUNS):
            # Query tree
            start = time()
            dist, idx = kdtree.query(QUERIES, K)
            qt = (time() - start)*1000
            query_time += qt
            
            # Print run results
            # print(f"Run {run}: {bt=:.3f} ms; {qt=:.3f} ms")

        avg_query = query_time / RUNS
        print(f"{lib} k={K}: {avg_query=:.3f} ms")
    print()

for lib in ["fnntw"]:
    print()
    print(f"{lib} single precision results (d={NRAND:,}, q={NQUERY:,}, ls={LS}, runs={RUNS})")
    DATA = DATA.astype(np.float32)
    for K in KK:
        query_time = 0
        build_time = 0
        kdtree = None
        # Generate queries
        QUERIES = np.random.uniform(size=(NQUERY, 3)).astype(np.float32)
        for run in range(RUNS):


            # Build tree
            start = time()
            kdtree = pyfnntw.Treef32(DATA, leafsize=LS, par_split_level=3)
            bt = (time() - start)*1000
            build_time += bt

            # Query tree
            start = time()
            dist, idx = kdtree.query(QUERIES, K)
            qt = (time() - start)*1000
            query_time += qt
            
            # Print run results
            # print(f"Run {run}: {bt=:.3f} ms; {qt=:.3f} ms")

        avg_build = build_time / RUNS
        avg_query = query_time / RUNS
        print(f"{lib} k={K}: {avg_build=:.3f} ms; {avg_query=:.3f} ms")
    print()


import pyfnntw;
import numpy as np;
from scipy.spatial import cKDTree as Tree
from time import time

ND = 10**3
NQ = 10**3
RUNS = 10
TRIALS = 10
WARMUP = 5

# Get some data and queries
data = np.random.uniform(size=(ND, 3))
query = np.random.uniform(size=(NQ, 3))

# Build and query the pynntw tree
tree1 = pyfnntw.Treef64(data, 32, 1)
(_, ids1) = tree1.query(query)

# Build and query the scipy tree
tree2 = Tree(data, 32)
(_, ids2) = tree2.query(query)

if np.all(ids1 == ids2):
    print("Success")
else:
    print("Failure")

# REPEAT WITH PERIODIC BOUNDARY CONDITIONS

# Get some data and queries
data = np.random.uniform(size=(ND, 3))
query = np.random.uniform(size=(NQ, 3))

# Build and query the pynntw tree
tree1 = pyfnntw.Treef64(data, 32, 1, boxsize=np.array([1.0, 1.0, 1.0]))
(_, ids1) = tree1.query(query)

# Build and query the scipy tree
tree2 = Tree(data, 32, boxsize=np.array([1.0, 1.0, 1.0]))
(_, ids2) = tree2.query(query)

if np.all(ids1 == ids2):
    print("Success")
else:
    print("Failure")



from time import perf_counter as time
import pyfnntw
import numpy as np
import scipy


RAND_QUERY = [
    (10**6, 10**5),
    (4*10**6, 4*10**6),
]
RUNS = 1
WARMUP = 0
LS = 32
K = 10

def get_trans_par(data, query, k):
    start = time()
    xtree = scipy.spatial.cKDTree(data, boxsize=1.0)
    bt = (time()-start)*1000

    start = time()
    _, disi = xtree.query(query, k=k)
    qt = (time()-start)*1000

    dis_trans = -np.ones((len(query), k))
    dis_par = -np.ones((len(query), k))
    for ik in range(k):
        ineighs = disi[:, ik]
        delta_pos = data[ineighs] - query
        delta_trans = np.sqrt(delta_pos[:, 0]**2 + delta_pos[:, 1]**2)
        delta_par = abs(delta_pos[:, 2])
        dis_trans[:, ik] = delta_trans          
        dis_par[:, ik] = delta_par

    return dis_trans, dis_par, bt, qt

def get_trans_par_fnntw(data, query, k):
    start = time()
    xtree = pyfnntw.Treef64(data, leafsize=LS, boxsize=np.array([1.0, 1.0, 1.0]))
    bt = (time()-start)*1000

    start = time()
    par, trans = xtree.query(query, k=k, axis=2)
    qt = (time()-start)*1000

    return trans, par, bt, qt

for (rand, query) in RAND_QUERY:
    for lib in ["scipy", "fnntw"]:
        query_time = 0
        build_time = 0
        total_time = 0
        bt = 0; qt = 0
        for run in range(-WARMUP, RUNS):
            DATA = np.random.uniform(size=(rand, 3))
            QUERIES = np.random.uniform(size=(query, 3))
            # Query tree
            if lib == "scipy":
                trans, par, bt, qt = get_trans_par(DATA, QUERIES, K)        
            elif lib == "fnntw":
                trans, par, bt, qt = get_trans_par_fnntw(DATA, QUERIES, K)
            if run >= 0:
                query_time += qt
                build_time += bt
                total_time += qt+bt
        avg_build = build_time / RUNS
        avg_query = query_time / RUNS
        avg_total = total_time / RUNS
        print(f"{lib:8}: build = {avg_build:.3f} ms; query = {avg_query:.3f} ms; total = {avg_total:.3f} ms")

    scipyt,scipyp,_,_= get_trans_par(DATA, QUERIES, K)   
    fnntwt,fnntwp,_,_= get_trans_par_fnntw(DATA, QUERIES, K)   
    print(f"{scipyt.shape=}; {scipyp.shape=}")
    print(f"{fnntwt.shape=}; {fnntwp.shape=}")
    print(f"{scipyt[:2]=};")
    print(f"{fnntwt[:2]=};")
    print(f"{scipyp[:2]=};")
    print(f"{fnntwp[:2]=};")



import numpy as np
from time import perf_counter as time
import pyfnntw
from pykdtree.kdtree import KDTree as pykdTree
from scipy.spatial import cKDTree as scipyTree


NRAND = 10**5
NQUERY = 10**6
WARMUP = 20
RUNS = 100
LS = 32
KK = [1, 10]

print(f"(d={NRAND:,}, q={NQUERY:,}, ls={LS}, runs={RUNS}")

# Generate data 
DATA = np.random.uniform(size=(NRAND, 3))

times = []
for _ in range((len(KK) + 1)):
    times.append([])
libs = ["scipy", "pykdtree", "fnntw"]
for lib in libs:
    print()
    print(f"{lib} results")
 
    build_time = 0
    kdtree = None
    # Generate queries
    for run in range(-WARMUP, RUNS):
        # Build tree
        start = time()
        if lib == "scipy":
            kdtree = scipyTree(DATA, leafsize=LS)#, boxsize = 1.0)
        elif lib == "pykdtree":
            kdtree = pykdTree(DATA, leafsize=LS)
        elif lib == "fnntw":
            kdtree = pyfnntw.Treef64(DATA, leafsize=LS, par_split_level=4)
        else:
            assert False
        bt = (time() - start)*1000
        if run >= 0:
            build_time += bt
    avg_build = build_time / RUNS
    times[0].append(avg_build)
    print(f"{lib}: {avg_build=:.3f} ms")

    for (i, K) in enumerate(KK):
        query_time = 0
        qt = 0
        queries = np.random.uniform(size=(NQUERY, 3))
        for run in range(-WARMUP, RUNS):
            # Query tree
            if lib == "scipy":
                start = time()
                dist, idx = kdtree.query(queries, k=K, workers=-1)        
                qt = (time() - start)*1000
            elif lib == "pykdtree":
                start = time()
                dist, idx = kdtree.query(queries, k=K, sqr_dists=False)
                qt = (time() - start)*1000
            elif lib == "fnntw":
                start = time()
                dist, idx = kdtree.query(queries, K)
                qt = (time() - start)*1000
            else:
                assert False
            if run >= 0:
                query_time += qt

        avg_query = query_time / RUNS
        times[i+1].append(avg_query)
        print(f"{lib} k={K}: {avg_query=:.3f} ms")
    print()
print('\x1b[6;30;42m' + 'build & nonpbc winners:' + '\x1b[0m')
print(f"build winner: {libs[np.argmin(times[0])]} @ {np.min(times[0]):.3f} ms")
for (i, K) in enumerate(KK):
    print(f"{K=} query winner: {libs[np.argmin(times[1+i])]}: @ {np.min(times[i+1]):.3f} ms")

print()
print("periodic results")
times = []
for _ in range(len(KK)):
    times.append([])
libs = ["scipy", "fnntw"]
for lib in libs:
    print()
    print(f"{lib} results")

    # Build tree (no need to benchmark)
    kdtree = None
    if lib == "scipy":
        kdtree = scipyTree(DATA, boxsize = 1.0, leafsize=LS)
    else:
        kdtree = pyfnntw.Treef64(DATA, boxsize = np.array([1.0]*3), leafsize=LS, par_split_level=4)

    for (i, K) in enumerate(KK):
        query_time = 0
        qt = 0
        queries = np.random.uniform(size=(NQUERY, 3))
        for run in range(-WARMUP, RUNS):
            # Query tree
            if lib == "scipy":
                start = time()
                dist, idx = kdtree.query(queries, k=K, workers=-1)        
                qt = (time() - start)*1000
            else:
                start = time()
                dist, idx = kdtree.query(queries, K)
                qt = (time() - start)*1000
            if run >= 0:
                query_time += qt

        avg_query = query_time / RUNS
        times[i].append(avg_query)
        print(f"{lib} k={K}: {avg_query=:.3f} ms")
    print()
print('\x1b[6;30;42m' + 'pbc winners:' + '\x1b[0m')
for (i, K) in enumerate(KK):
    print(f"{K=} query winner: {libs[np.argmin(times[i])]} @ {np.min(times[i]):.3f} ms", )


times = []
for _ in range(len(KK)):
    times.append([])
libs = ["scipy", "pykdtree", "fnntw"]
DATA = DATA.astype(np.float32)
for lib in libs:
    print()
    print(f"{lib} single precision results")

    build_time = 0
    kdtree = None
    # Generate queries
    queries = np.random.uniform(size=(NQUERY, 3)).astype(np.float32)
    # Build tree (no need to benchmark)
    if lib == "scipy":
        kdtree = scipyTree(DATA, leafsize=LS)#, boxsize = 1.0)
    elif lib == "pykdtree":
        kdtree = pykdTree(DATA, leafsize=LS)
    else:
        kdtree = pyfnntw.Treef32(DATA, leafsize=LS, par_split_level=4)
        
    for (i, K) in enumerate(KK):
        query_time = 0
        qt = 0
        queries = np.random.uniform(size=(NQUERY, 3)).astype(np.float32)
        for run in range(-WARMUP, RUNS):
            # Query tree
            if lib == "scipy":
                start = time()
                dist, idx = kdtree.query(queries, k=K, workers=-1)        
                qt = (time() - start)*1000
            elif lib == "pykdtree":
                start = time()
                dist, idx = kdtree.query(queries, k=K, sqr_dists=False)
                qt = (time() - start)*1000
            else:
                start = time()
                dist, idx = kdtree.query(queries, K)
                qt = (time() - start)*1000
            if run >= 0:
                query_time += qt
            
        avg_query = query_time / RUNS
        times[i].append(avg_query)
        print(f"{lib} k={K}: {avg_query=:.3f} ms")
    print()
print('\x1b[6;30;42m' + 'single precision winners:' + '\x1b[0m')
for (i, K) in enumerate(KK):
    print(f"{K=} query winner: {libs[np.argmin(times[i])]}  @ {np.min(times[i]):.3f} ms", )
